// early version IR project for bus pirate. Expect enhancements over time

// Simple use of BusPirate gadget to send Infrared Remote messages
// amateur code. Author accepts no responsibility for anything, anywhere, ever
// use at your own risk

// Bus Pirate uses PIC24FJ64GA002 - I have an earlier version board
// No crystal fitted, so use internal oscillator set to 8MHz PLLx4 gives 32MHz clock

// program bus pirate using picKIT 3 or similar - not designed for bootloader
// connect IR LED (driven by transistor) to AUX1 pin 10 (RB10)
// bus pirate provides 3.3V and 5V for LED driver circuit if wanted
// use terminal program 9600 baud, 8, n, 1 - USB port

// note that most numbers are in hexadecimal format
// parameters are processed as longwords, but for this program we only use them as bytes

// ideas for enhancement:
//    other IR protocols
//    UART function ?
//    monitor ADC inputs ?

// PIC24FJ64GA002 Pin info:
//  2 RA0  VREGEN - take high to enable 3.3V and 5V to output plug
//  3 RA1  LD1 - Red LED
// 10 RA3  CTS~ for FT232RLSSOP pin 11
// 11 RB4  RX       FT232 pin 5
// 12 RA4  RTS~     FT232 pin 3
// 14 RB5  TX       FT232 pin 1
// 15 RB6  CS       IO pin 2
// 16 RB7  MISO     IO pin 1
// 17 RB8  CLK      IO pin 4
// 18 RB9  MOSI     IO pin 3
// 21 RB10 AUX1     IO pin 5
// 22 RB11 PULLUPEN output a high to enable 3.3V and 5V supplys to IO
// 23 RB12 ADC1     (IO pin 7) via 2x10k divider
// 24 RB13 ADC2     (IO pin 6) via 2x10k divider to measure VEXTERN
// 25 RB14 ADC3     (IO pin 9) via 2x10k divider to measure 3.3V switchable
// 26 RB15 ADC4     (IO pin 8) via 2x10k divider to measure   5V switchable

// bus pirate Connector pins - pin 9 is closest to corner of board - brackets show non-direct connection
//  1  RB7   MISO
//  2  RB6   CS
//  3  RB9   MOSI
//  4  RB8   CLK
//  5  RB10  AUX1
//  6 (RB13) VEXTERN
//  7 (RB12) ADC
//  8 (RB15) SWV50  (+5 Volts)   switchable
//  9 (RB14) SWV33  (+3.3 Volts) switchable
// 10  GND

// LEDs:
// Red     5V supply from USB (nearest USB Skt)
// Blue    Rx driven by FT232 pin 22 - Host Rx, PIC Tx
// Green   Tx driven by FT232 pin 23 - Host TX, PIC Rx
// Yellow  (RA0) 3.3 Volts is on (5V too)

// pin definitions
sbit VREGENABLE at LATAbits.B0;                                                 // Turn on 3.3V and 5V and yellow LED
sbit LD1 at LATAbits.B1;                                                        // red LED
sbit ir_out at LATBbits.B10;                                                    // output to IR LED  (AUX1)

// prototypes
void command_n(void);                                                           // send NEC IR message
void linein(char* pstring);                                                     // get a line of text characters
char getparams(char* tempstring);                                               // decode line of text into command and up to 8 parameters
void PrintHex(unsigned char i);                                                 // print a char as hex
void helptext(void);
void NEC_zero();
void NEC_one();
void IR_mark(unsigned char n_cycles);
void IR_space(unsigned char n_cycles);
void NEC_byte_out(unsigned char send_char);
void NEC_out(void);
void NEC_repeat(void);

// global variables
unsigned char inbuff[257];                                                      // 256 byte buffer for input string + 1 for null
unsigned char cmd;                                                              // first character of command
unsigned char param[8][20];                                                     // params 0 to 7 as strings
unsigned long upar[8];                                                          // params as longs
unsigned int nparams;                                                           // count number of params
unsigned debug;                                                                 // sets level of debug info output

// main program
void main() {
signed char ctr;                                                                // just a loop counter

    Delay_ms(2000);                                                             // prevent multi-start when programming
    debug = 0;                                                                  // set default amount of debug info
    CLKDIV = 0x0000;                                                            // 8MHz internal oscillator (fed to 4xPLL)
    AD1PCFG = 0xffff;                                                           // all digital
    TRISA = 0x0000;
    TRISB = 0x0020;                                                             // bit 5 Rx in, rest output
    LATA = 0;
    LATB = 0;
    VREGENABLE = 1;                                                             // turn on 3.3V and 5V
    for( ctr = 0 ; ctr < 8 ; ctr++){
        upar[ctr] = 0UL;                                                        // initialise parameters
    }
    
    PPS_Mapping_NoLock(4, _OUTPUT, _U2TX);                                      // Sets RP4 to be Output, and maps EUSART2 Asynchronous Transmit/Synchronous Clock Output to it
    PPS_Mapping_NoLock(5, _INPUT, _U2RX);                                       // Sets RP5 to be Input, and maps RX2/DT2 Input to it

    UART2_Init(9600);                                                           // Initialize UART module at 9600 bps
    Delay_ms(100);                                                              // Wait for UART module to stabilize

    helptext();                                                                 // display initial help text

    while(1){
        UART2_Write_Text("\r\n>");
        linein(inbuff);                                                         // wait for a line of text
        getparams(inbuff);                                                      // parse the text
        
        if(debug > 0){                                                          // check if debug info wanted
            UART2_Write_Text("\r\nNumber of parameters = ");
            PrintHex(nparams);
            UART2_Write_Text("\r\nP0 = ");                                      // parameter 0 is command letter
            UART2_Write(cmd);
            UART2_Write_Text(" (Command)");
            for( ctr = 1; ctr < 8 ; ctr++){                                     // remaining 7 parameters
                UART2_Write_Text("\r\nP");
                UART2_Write(ctr + 48);                                          // write number as ASCII
                UART2_Write_Text(" = ");
                PrintHex(upar[ctr]);
            }
        }

        switch(cmd){
            case '?':
                helptext();
                break;
            case 'D':
                if(nparams > 1){
                    debug = upar[1];
                }
                break;
            case 'N':
                command_n();
                break;
            default:
                UART2_Write_Text("command not recognised\r\n");
                break;
        }
    }
}

// command N - send NEC IR message
void command_n(void){
    UART2_Write_Text("\r\nsend NEC IR message");
    NEC_out();
}

// Extended NEC protocol: 1/4 or 1/3 duty cycle 38khz (21 pulses per bit approx)
// 9ms mark then 4.5ms space	header
// 560us mark, 560us space	zero
// 560us mark, 1.69ms space	one
// 4 bytes, LSB first
// 560us mark - long space
// stop bit of one
// option 1 = repeat every 110ms (if wanted), but just 9ms mark, 2.25 space, 560us mark (no data)
// option 2 = repeat full message
// 00ff = default address, custom codes available, byte 4 is inverse of byte 3
// byte1 is low byte of address, byte2 is either inverse of byte 1 or high byte

// Send NEC IR message
// use parameters 1,2,3 and invert byte 3 to get byte 4
// parameter 4 is number of repeat messages to send (max 15)
void NEC_out(void){
signed char ctr;

    IR_mark(173);                                                               // 9mS mark is 346 cycles of 38 kHz
    IR_mark(173);                                                               // but function only takes byte parameter
    IR_space(173);                                                              // 4.5mS space is 173 cycles of 38 kHz
    NEC_byte_out(upar[1]);
    NEC_byte_out(upar[2]);
    NEC_byte_out(upar[3]);
    NEC_byte_out(~upar[3]);                                                     // inverse of byte 3
    NEC_one();                                                                  // stop bit of one
    
    ctr = (upar[4] & 0x0f);                                                     // maximum of 15 repeats
    while(ctr){                                                                 // parameter 4 is number of repeats
        NEC_repeat();                                                           // simple repeat sequence
        ctr--;
    }
}

// repeat every 110ms (if wanted), but just 9ms mark, 2.25 space, 560us mark (no data)
void NEC_repeat(void){
    Delay_ms(38);                                                               // must have a no-data period
    IR_mark(173);                                                               // 9mS mark is 346 cycles of 38 kHz
    IR_mark(173);                                                               // but function only takes byte parameter
    IR_space(87);                                                               // 2.25mS space is 87 cycles of 38 kHz
    NEC_one();                                                                  // stop bit of one
}

// print help text
void helptext(void){
    UART2_Write_Text("\r\n?            help - 05 Nov 2018");
    UART2_Write_Text("\r\n             numbers in hex");
    UART2_Write_Text("\r\n             omit number for no change\r\n");
    UART2_Write_Text("\r\nd a          set debug level to a");
    UART2_Write_Text("\r\nn aa aa dd r send NEC IR message hex");
    UART2_Write_Text("\r\n");
}

// UART2 write text and new line (carriage return + line feed)
void UART2_Write_Line(char *uart_text) {
    UART2_Write_Text(uart_text);
    UART2_Write_Text("\r\n");
}

// display byte in hex
void PrintHex(unsigned char i) {
unsigned char hi,lo;

    hi = i & 0xF0;                                                              // high nibble
    hi = hi >> 4;
    hi = hi + '0';
    if (hi > '9') hi = hi + 7;
    lo = (i & 0x0F) + '0';                                                      // low nibble
    if (lo > '9') lo = lo + 7;

    UART2_Write(hi);
    UART2_Write(lo);
    UART2_Write(' ');
}

// get parameters from string
char getparams(char* tempstring){
char oldcmd;                                                                    // remember old command letter
int n;                                                                          // counter
unsigned char tempchar;                                                         // temp store
unsigned char tempchar2;                                                        // temp store
unsigned int pointer1 = 0;                                                      // pointer within string
unsigned int pointer2 = 0;                                                      // pointer within string

    oldcmd = cmd;                                                               // remember old value
    nparams = 0;                                                                // initialise
    pointer1 = 0;                                                               // initialise
    
    for ( n = 0 ; n < 9 ; n++){                                                 // count 0 to 7

        /* skip leading spaces */
        tempchar=1;                                                             // init
        do {
            tempchar = tempstring[pointer1++];
        } while (tempchar == 0x20);
        pointer1--;                                                             // leave pointer after param

        // count parameters
        if ((tempchar != 0) && (nparams < 8)){
            nparams++;
        }

        if (n == 0){                                                            // only for param 0
            tempchar = tempstring[pointer1];                                    // get only first letter
            if ((tempchar > 0x60) && (tempchar < 0x7b)) {
                tempchar -= 0x20;                                               // to upper case
            }
            cmd=tempchar;
        }

        tempchar=1;                                                             // initialise
        if (nparams != 0){                                                      // only for non-blank line
            // get all characters
            pointer2=0;                                                         // initialise
            do {
                tempchar = tempstring[pointer1++];                              // get char
                if ((tempchar > 0x60) && (tempchar < 0x7b)) {
                    tempchar -= 0x20;                                           // to upper case
                }
                param[n][pointer2++] = tempchar;                                // and store it
            } while ((tempchar > 0x20) && (tempchar < 0x7f) && (pointer2 < 20));   // text
            param[n][pointer2-1] = 0;                                           // zero terminate param string
            if((tempchar == 0) || (n ==9 )){                                    // no more params if end of line
                break;
            }
        }
    }

    // convert parameter strings to hex longwords
    for (n = 0; n < nparams ; n++){                                             // parms are 0 to 7

        tempchar2 = 1;                                                          // initialise
        pointer1 = 0;                                                           // initialise
        upar[n] = 0;                                                            // initialise
        do {
            tempchar = param[n][pointer1++];                                    // read one character
            if ( tempchar != 0 ){
                upar[n] = upar[n] * 0x10;                                       // next most signicant hex digit
            }
            tempchar2 = tempchar;                                               // remember original character
            if ((tempchar < 0x30) || (tempchar > 0x46)) {
                tempchar = 0x00;                                                // non-hex to zero
            }
            if ((tempchar > 0x39) && (tempchar < 0x41)) {
                tempchar = 0x00;                                                // non-hex to zero
            }
            if ((tempchar > 0x40) && (tempchar < 0x47)) {
                tempchar -= 0x37;                                               // to bin
            }
            if ((tempchar > 0x2f) && (tempchar < 0x3a)) {
                tempchar -= 0x30;                                               // to bin
            }
                upar[n] += tempchar;
        } while (tempchar2 != 0);

    }

    if (nparams == 0) {
        cmd = oldcmd;
    }
    return cmd;
}

// This function gets a string from UART2.
void linein(char* pstring){
char inChar=0;

    while(inChar != '\r'){

        if (UART2_Data_Ready()){                                                //UART read
            inChar = UART2_Read();
            UART2_Write(inChar);
            if(inChar == '\r'){
                UART2_Write('\n');
                *pstring++ = '\0';
            }
            else {
                if (inChar != 0x7f ){                                           // if not backspace
                    *pstring++ = inChar;
                }
                if (inChar == 0x7f ){                                           //handle backspace
                    *--pstring = 0x00;
                }
            }
        }// End of if (UART"IsPressed()){...
    }// end of while..
}

// read a single charater from UART2
unsigned char chin(void){
unsigned char temp;

    while(!UART2_Data_Ready());
    temp=UART2_Read();
    return temp;
}

// write a single chc to UART2 and add lf to cr
void chout(unsigned char outchar){

    UART2_Write(outchar);
    if(outchar == 0x0d){
        UART2_Write(0x0a);
    }
}

// send a single byte of data to IR in NEC format, but modulated with approx 38kHz
// 8 bits, rate set by IR_one and IR_zero functions
void NEC_byte_out(unsigned char send_char){
unsigned char n;                                                                // just a counter

    for(n=0 ; n<8 ; n++){                                                       // count through 8 bits, bit 0 sent first
        if(send_char & (1 << n)){                                               // check bit(n) value
            NEC_one();                                                           // send a one to IR
        }
        else{
            NEC_zero();                                                          // send a zero to IR
    }   }
}

// send an NEC one bit to IR - 560uS mark, 1680uS space
// note that spaces get stretched in receiver - so received timings will not be exact
void NEC_one(){
    IR_mark(21);
    IR_space(63);                                                               // adjust to give correct rate
}

// send an NEC zero bit to IR - 560uS mark, 560uS space
// note that marks get shortened in receiver - so received timings will not be exact
void NEC_zero(){                                                                // adjust to give correct rate
    IR_mark(21);
    IR_space(21);                                                               //
}

// output n-cycles of 38 kHz to IR LED on GP0 (0)
// note that timings seem off and the mark-space ratio is less than 50%
//    ... this is because loop/call times get added to the total.
// real IR signals tend to be 25 to 33% duty cycle, so less than 50% is not a problem
void IR_mark(unsigned char n_cycles){
    while(n_cycles){
        ir_out = 1;                                // high to turn on IR LED
        Delay_us(13);                              // half of 38kHz square wave
        ir_out = 0;                                // low to turn off IR LED
        Delay_us(13);                              // half of 38kHz square wave
        n_cycles--;
    }
}

// output n_cycles of nothing
// - no I am not mad - copying the above is an easy way to get the timings right
void IR_space(unsigned char n_cycles){
    while(n_cycles){
        ir_out = 0;                                // low to turn off IR LED
        Delay_us(13);                              // half of 38kHz square wave
        ir_out = 0;                                // low to turn off IR LED
        Delay_us(13);                              // half of 38kHz square wave
        n_cycles--;
    }
}
